@@ -34,7 +34,7 @@ import java.util.Random;
  * @author Rodney Waldhoff
  * @author Dirk Verbeeck
  * @author Sandy McArthur
- * @version $Revision: 609415 $ $Date: 2008-01-06 14:45:32 -0700 (Sun, 06 Jan 2008) $
+ * @version $Revision: 791860 $ $Date: 2009-07-07 11:10:30 -0400 (Tue, 07 Jul 2009) $
  */
 public class TestGenericObjectPool extends TestBaseObjectPool {
     public TestGenericObjectPool(String testName) {
@@ -67,6 +67,7 @@ public class TestGenericObjectPool extends TestBaseObjectPool {
 
     public void tearDown() throws Exception {
         super.tearDown();
+        pool.clear();
         pool.close();
         pool = null;
     }
@@ -95,6 +96,7 @@ public class TestGenericObjectPool extends TestBaseObjectPool {
             // expected
         }
         pool.returnObject(obj1);
+        assertEquals(1, pool.getNumIdle());
         pool.close();
     }
 
@@ -131,7 +133,7 @@ public class TestGenericObjectPool extends TestBaseObjectPool {
         GenericObjectPool pool = new GenericObjectPool(factory);
         pool.setMaxActive(2);
         pool.setMinIdle(1);
-        Object obj = pool.borrowObject(); // numActive = 1, numIdle = 0
+        pool.borrowObject(); // numActive = 1, numIdle = 0
         // Create a test thread that will run once and try a borrow after
         // 150ms fixed delay
         TestThread borrower = new TestThread(pool, 1, 150, false);
@@ -291,7 +293,7 @@ public class TestGenericObjectPool extends TestBaseObjectPool {
         pool.evict();
         // Should hit 6,7,0 - 0 for second time
         for (int i = 0; i < 8; i++) {
-            VisitTracker tracker = (VisitTracker) pool.borrowObject();    
+            VisitTracker tracker = (VisitTracker) pool.borrowObject();
             if (tracker.getId() != 0) {
                 assertEquals("Instance " +  tracker.getId() + 
                         " visited wrong number of times.",
@@ -307,12 +309,19 @@ public class TestGenericObjectPool extends TestBaseObjectPool {
         int[] smallPrimes = {2, 3, 5, 7};
         Random random = new Random();
         random.setSeed(System.currentTimeMillis());
-        pool.setMaxIdle(-1);
         for (int i = 0; i < 4; i++) {
             pool.setNumTestsPerEvictionRun(smallPrimes[i]);
             for (int j = 0; j < 5; j++) {
-                pool.clear();
+                pool = new GenericObjectPool(factory);
+                pool.setNumTestsPerEvictionRun(3);
+                pool.setMinEvictableIdleTimeMillis(-1);
+                pool.setTestWhileIdle(true);
+                pool.setLifo(lifo);
+                pool.setTestOnReturn(false);
+                pool.setTestOnBorrow(false);
+                pool.setMaxIdle(-1);
                 int instanceCount = 10 + random.nextInt(20);
+                pool.setMaxActive(instanceCount);
                 for (int k = 0; k < instanceCount; k++) {
                     pool.addObject();
                 }
@@ -333,6 +342,7 @@ public class TestGenericObjectPool extends TestBaseObjectPool {
                 int visitCount = 0;
                 for (int k = 0; k < instanceCount; k++) {
                     tracker = (VisitTracker) pool.borrowObject(); 
+                    assertTrue(pool.getNumActive() <= pool.getMaxActive());
                     visitCount = tracker.getValidateCount();                  
                     assertTrue(visitCount >= cycleCount && 
                             visitCount <= cycleCount + 1);
@@ -350,6 +360,65 @@ public class TestGenericObjectPool extends TestBaseObjectPool {
         assertEquals(0,pool.getNumIdle());
         pool.close();
     }
+    
+    public void testExceptionOnDestroyDuringBorrow() throws Exception {
+        SimpleFactory factory = new SimpleFactory(); 
+        factory.setThrowExceptionOnDestroy(true);
+        GenericObjectPool pool = new GenericObjectPool(factory);
+        pool.setTestOnBorrow(true);
+        pool.borrowObject();
+        factory.setValid(false); // Make validation fail on next borrow attempt
+        try {
+            pool.borrowObject();
+            fail("Expecting NoSuchElementException");
+        } catch (NoSuchElementException ex) {
+            // expected
+        }
+        assertEquals(1, pool.getNumActive());
+        assertEquals(0, pool.getNumIdle());
+    }
+    
+    public void testExceptionOnDestroyDuringReturn() throws Exception {
+        SimpleFactory factory = new SimpleFactory(); 
+        factory.setThrowExceptionOnDestroy(true);
+        GenericObjectPool pool = new GenericObjectPool(factory);
+        pool.setTestOnReturn(true);
+        Object obj1 = pool.borrowObject();
+        pool.borrowObject();
+        factory.setValid(false); // Make validation fail
+        pool.returnObject(obj1);
+        assertEquals(1, pool.getNumActive());
+        assertEquals(0, pool.getNumIdle());
+    }
+    
+    public void testExceptionOnActivateDuringBorrow() throws Exception {
+        SimpleFactory factory = new SimpleFactory(); 
+        GenericObjectPool pool = new GenericObjectPool(factory);
+        Object obj1 = pool.borrowObject();
+        Object obj2 = pool.borrowObject();
+        pool.returnObject(obj1);
+        pool.returnObject(obj2);
+        factory.setThrowExceptionOnActivate(true);
+        factory.setEvenValid(false);  
+        // Activation will now throw every other time
+        // First attempt throws, but loop continues and second succeeds
+        Object obj = pool.borrowObject();
+        assertEquals(1, pool.getNumActive());
+        assertEquals(0, pool.getNumIdle());
+        
+        pool.returnObject(obj);
+        factory.setValid(false);
+        // Validation will now fail on activation when borrowObject returns
+        // an idle instance, and then when attempting to create a new instance
+        try {
+            obj1 = pool.borrowObject();
+            fail("Expecting NoSuchElementException");
+        } catch (NoSuchElementException ex) {
+            // expected
+        }
+        assertEquals(0, pool.getNumActive());
+        assertEquals(0, pool.getNumIdle());
+    }
 
     public void testSetFactoryWithActiveObjects() throws Exception {
         GenericObjectPool pool = new GenericObjectPool();
@@ -436,6 +505,25 @@ public class TestGenericObjectPool extends TestBaseObjectPool {
             // expected
         }
     }
+    
+    public void testTimeoutNoLeak() throws Exception {
+        pool.setMaxActive(2);
+        pool.setMaxWait(10);
+        pool.setWhenExhaustedAction(GenericObjectPool.WHEN_EXHAUSTED_BLOCK);
+        Object obj = pool.borrowObject();
+        Object obj2 = pool.borrowObject();
+        try {
+            pool.borrowObject();
+            fail("Expecting NoSuchElementException");
+        } catch (NoSuchElementException ex) {
+            //xpected
+        }
+        pool.returnObject(obj2);
+        pool.returnObject(obj);
+        
+        obj = pool.borrowObject();
+        obj2 = pool.borrowObject();
+    }
 
     public void testMaxActiveZero() throws Exception {
         pool.setMaxActive(0);
@@ -449,6 +537,84 @@ public class TestGenericObjectPool extends TestBaseObjectPool {
         }
     }
 
+    public void testMaxActiveUnderLoad() {
+        // Config
+        int numThreads = 199; // And main thread makes a round 200.
+        int numIter = 20;
+        int delay = 25;
+        int maxActive = 10;
+        
+        SimpleFactory factory = new SimpleFactory();
+        factory.setMaxActive(maxActive);
+        pool.setFactory(factory);
+        pool.setMaxActive(maxActive);
+        pool.setWhenExhaustedAction(GenericObjectPool.WHEN_EXHAUSTED_BLOCK);
+        pool.setTimeBetweenEvictionRunsMillis(-1);
+        
+        // Start threads to borrow objects
+        TestThread[] threads = new TestThread[numThreads];
+        for(int i=0;i<numThreads;i++) {
+            // Factor of 2 on iterations so main thread does work whilst other
+            // threads are running. Factor of 2 on delay so average delay for
+            // other threads == actual delay for main thread
+            threads[i] = new TestThread(pool, numIter * 2, delay * 2);
+            Thread t = new Thread(threads[i]);
+            t.start();
+        }
+        // Give the threads a chance to start doing some work
+        try {
+            Thread.sleep(5000);
+        } catch(InterruptedException e) {
+            // ignored
+        }
+        
+        for (int i = 0; i < numIter; i++) {
+            Object obj = null;
+            try {
+                try {
+                    Thread.sleep(delay);
+                } catch(InterruptedException e) {
+                    // ignored
+                }
+                obj = pool.borrowObject();
+                // Under load, observed _numActive > _maxActive 
+                if (pool.getNumActive() > pool.getMaxActive()) {
+                    throw new IllegalStateException("Too many active objects");
+                }
+                try {
+                    Thread.sleep(delay);
+                } catch(InterruptedException e) {
+                    // ignored
+                }
+            } catch (Exception e) {
+                // Shouldn't happen
+                e.printStackTrace();
+                fail("Exception on borrow");
+            } finally {
+                if (obj != null) {
+                    try {
+                        pool.returnObject(obj);
+                    } catch (Exception e) {
+                        // Ignore
+                    }
+                }
+            }
+        }
+
+        for(int i=0;i<numThreads;i++) {
+            while(!(threads[i]).complete()) {
+                try {
+                    Thread.sleep(500L);
+                } catch(InterruptedException e) {
+                    // ignored
+                }
+            }
+            if(threads[i].failed()) {
+                fail("Thread "+i+" failed: "+threads[i]._error.toString());
+            }
+        }
+    }
+
     public void testInvalidWhenExhaustedAction() throws Exception {
         try {
             pool.setWhenExhaustedAction(Byte.MAX_VALUE);
@@ -699,7 +865,7 @@ public class TestGenericObjectPool extends TestBaseObjectPool {
             pool.setTimeBetweenEvictionRunsMillis(50L);
             
             // wait a second (well, .2 seconds)
-            try { Thread.sleep(200L); } catch(Exception e) { }
+            try { Thread.sleep(200L); } catch(InterruptedException e) { }
             
             // assert that the evictor has cleared out the pool
             assertEquals("Should have 0 idle",0,pool.getNumIdle());
@@ -725,13 +891,13 @@ public class TestGenericObjectPool extends TestBaseObjectPool {
             pool.returnObject(active[i]);
         }
 
-        try { Thread.sleep(100L); } catch(Exception e) { }
+        try { Thread.sleep(100L); } catch(InterruptedException e) { }
         assertTrue("Should at most 6 idle, found " + pool.getNumIdle(),pool.getNumIdle() <= 6);
-        try { Thread.sleep(100L); } catch(Exception e) { }
+        try { Thread.sleep(100L); } catch(InterruptedException e) { }
         assertTrue("Should at most 3 idle, found " + pool.getNumIdle(),pool.getNumIdle() <= 3);
-        try { Thread.sleep(100L); } catch(Exception e) { }
+        try { Thread.sleep(100L); } catch(InterruptedException e) { }
         assertTrue("Should be at most 2 idle, found " + pool.getNumIdle(),pool.getNumIdle() <= 2);
-        try { Thread.sleep(100L); } catch(Exception e) { }
+        try { Thread.sleep(100L); } catch(InterruptedException e) { }
         assertEquals("Should be zero idle, found " + pool.getNumIdle(),0,pool.getNumIdle());
     }
 
@@ -751,17 +917,17 @@ public class TestGenericObjectPool extends TestBaseObjectPool {
             pool.returnObject(active[i]);
         }
 
-        try { Thread.sleep(1000L); } catch(Exception e) { }
+        try { Thread.sleep(1000L); } catch(InterruptedException e) { }
         assertTrue("Should be less than 500 idle, found " + pool.getNumIdle(),pool.getNumIdle() < 500);
-        try { Thread.sleep(600L); } catch(Exception e) { }
+        try { Thread.sleep(600L); } catch(InterruptedException e) { }
         assertTrue("Should be less than 400 idle, found " + pool.getNumIdle(),pool.getNumIdle() < 400);
-        try { Thread.sleep(600L); } catch(Exception e) { }
+        try { Thread.sleep(600L); } catch(InterruptedException e) { }
         assertTrue("Should be less than 300 idle, found " + pool.getNumIdle(),pool.getNumIdle() < 300);
-        try { Thread.sleep(600L); } catch(Exception e) { }
+        try { Thread.sleep(600L); } catch(InterruptedException e) { }
         assertTrue("Should be less than 200 idle, found " + pool.getNumIdle(),pool.getNumIdle() < 200);
-        try { Thread.sleep(600L); } catch(Exception e) { }
+        try { Thread.sleep(600L); } catch(InterruptedException e) { }
         assertTrue("Should be less than 100 idle, found " + pool.getNumIdle(),pool.getNumIdle() < 100);
-        try { Thread.sleep(600L); } catch(Exception e) { }
+        try { Thread.sleep(600L); } catch(InterruptedException e) { }
         assertEquals("Should be zero idle, found " + pool.getNumIdle(),0,pool.getNumIdle());
 
         for(int i=0;i<500;i++) {
@@ -771,27 +937,21 @@ public class TestGenericObjectPool extends TestBaseObjectPool {
             pool.returnObject(active[i]);
         }
 
-        try { Thread.sleep(1000L); } catch(Exception e) { }
+        try { Thread.sleep(1000L); } catch(InterruptedException e) { }
         assertTrue("Should be less than 500 idle, found " + pool.getNumIdle(),pool.getNumIdle() < 500);
-        try { Thread.sleep(600L); } catch(Exception e) { }
+        try { Thread.sleep(600L); } catch(InterruptedException e) { }
         assertTrue("Should be less than 400 idle, found " + pool.getNumIdle(),pool.getNumIdle() < 400);
-        try { Thread.sleep(600L); } catch(Exception e) { }
+        try { Thread.sleep(600L); } catch(InterruptedException e) { }
         assertTrue("Should be less than 300 idle, found " + pool.getNumIdle(),pool.getNumIdle() < 300);
-        try { Thread.sleep(600L); } catch(Exception e) { }
+        try { Thread.sleep(600L); } catch(InterruptedException e) { }
         assertTrue("Should be less than 200 idle, found " + pool.getNumIdle(),pool.getNumIdle() < 200);
-        try { Thread.sleep(600L); } catch(Exception e) { }
+        try { Thread.sleep(600L); } catch(InterruptedException e) { }
         assertTrue("Should be less than 100 idle, found " + pool.getNumIdle(),pool.getNumIdle() < 100);
-        try { Thread.sleep(600L); } catch(Exception e) { }
+        try { Thread.sleep(600L); } catch(InterruptedException e) { }
         assertEquals("Should be zero idle, found " + pool.getNumIdle(),0,pool.getNumIdle());
     }
-
-    /**
-     * This test has a number of problems.
-     * 1. without the wait code in the for loop the create time for each instance
-     * was usually the same due to clock presision.
-     * 2. It's very hard to follow.
-     */
-    public void DISABLEDtestEvictionSoftMinIdle() throws Exception {
+ 
+    public void testEvictionSoftMinIdle() throws Exception {
         GenericObjectPool pool = null;
         
         class TimeTest extends BasePoolableObjectFactory {
@@ -813,37 +973,29 @@ public class TestGenericObjectPool extends TestBaseObjectPool {
         pool.setMaxActive(5);
         pool.setNumTestsPerEvictionRun(5);
         pool.setMinEvictableIdleTimeMillis(3000L);
-        pool.setTimeBetweenEvictionRunsMillis(250L);
-        pool.setTestWhileIdle(true);
         pool.setSoftMinEvictableIdleTimeMillis(1000L);
         pool.setMinIdle(2);
-        
+
         Object[] active = new Object[5];
         Long[] creationTime = new Long[5] ;
-        long lastCreationTime = System.currentTimeMillis();
         for(int i=0;i<5;i++) {
-            // make sure each instance has a different currentTimeMillis()
-            while (lastCreationTime == System.currentTimeMillis()) {
-                Thread.sleep(1);
-            }
             active[i] = pool.borrowObject();
             creationTime[i] = new Long(((TimeTest)active[i]).getCreateTime());
-            lastCreationTime = creationTime[i].longValue();
         }
         
         for(int i=0;i<5;i++) {
             pool.returnObject(active[i]);
         }
-        
-        try { Thread.sleep(1500L); } catch(Exception e) { }
-        assertTrue("Should be 2 OLD idle, found " + pool.getNumIdle(),pool.getNumIdle() == 2 &&
-                ((TimeTest)pool.borrowObject()).getCreateTime() == creationTime[3].longValue() &&
-                ((TimeTest)pool.borrowObject()).getCreateTime() == creationTime[4].longValue());
-        
-        try { Thread.sleep(2000L); } catch(Exception e) { }
-        assertTrue("Should be 2 NEW idle , found " + pool.getNumIdle(),pool.getNumIdle() == 2 &&
-                ((TimeTest)pool.borrowObject()).getCreateTime() != creationTime[0].longValue() &&
-                ((TimeTest)pool.borrowObject()).getCreateTime() != creationTime[1].longValue());
+
+        // Soft evict all but minIdle(2)
+        Thread.sleep(1500L);
+        pool.evict();
+        assertEquals("Idle count different than expected.", 2, pool.getNumIdle());
+
+        // Hard evict the rest.
+        Thread.sleep(2000L);
+        pool.evict();
+        assertEquals("Idle count different than expected.", 0, pool.getNumIdle());
     }
 
     public void testMinIdle() throws Exception {
@@ -855,27 +1007,27 @@ public class TestGenericObjectPool extends TestBaseObjectPool {
         pool.setTimeBetweenEvictionRunsMillis(100L);
         pool.setTestWhileIdle(true);
 
-        try { Thread.sleep(150L); } catch(Exception e) { }
+        try { Thread.sleep(150L); } catch(InterruptedException e) { }
         assertTrue("Should be 5 idle, found " + pool.getNumIdle(),pool.getNumIdle() == 5);
 
         Object[] active = new Object[5];
         active[0] = pool.borrowObject();
 
-        try { Thread.sleep(150L); } catch(Exception e) { }
+        try { Thread.sleep(150L); } catch(InterruptedException e) { }
         assertTrue("Should be 5 idle, found " + pool.getNumIdle(),pool.getNumIdle() == 5);
 
         for(int i=1 ; i<5 ; i++) {
             active[i] = pool.borrowObject();
         }
 
-        try { Thread.sleep(150L); } catch(Exception e) { }
+        try { Thread.sleep(150L); } catch(InterruptedException e) { }
         assertTrue("Should be 5 idle, found " + pool.getNumIdle(),pool.getNumIdle() == 5);
 
         for(int i=0 ; i<5 ; i++) {
             pool.returnObject(active[i]);
         }
 
-        try { Thread.sleep(150L); } catch(Exception e) { }
+        try { Thread.sleep(150L); } catch(InterruptedException e) { }
         assertTrue("Should be 10 idle, found " + pool.getNumIdle(),pool.getNumIdle() == 10);
     }
 
@@ -888,40 +1040,40 @@ public class TestGenericObjectPool extends TestBaseObjectPool {
         pool.setTimeBetweenEvictionRunsMillis(100L);
         pool.setTestWhileIdle(true);
 
-        try { Thread.sleep(150L); } catch(Exception e) { }
+        try { Thread.sleep(150L); } catch(InterruptedException e) { }
         assertTrue("Should be 5 idle, found " + pool.getNumIdle(),pool.getNumIdle() == 5);
 
         Object[] active = new Object[10];
 
-        try { Thread.sleep(150L); } catch(Exception e) { }
+        try { Thread.sleep(150L); } catch(InterruptedException e) { }
         assertTrue("Should be 5 idle, found " + pool.getNumIdle(),pool.getNumIdle() == 5);
 
         for(int i=0 ; i<5 ; i++) {
             active[i] = pool.borrowObject();
         }
 
-        try { Thread.sleep(150L); } catch(Exception e) { }
+        try { Thread.sleep(150L); } catch(InterruptedException e) { }
         assertTrue("Should be 5 idle, found " + pool.getNumIdle(),pool.getNumIdle() == 5);
 
         for(int i=0 ; i<5 ; i++) {
             pool.returnObject(active[i]);
         }
 
-        try { Thread.sleep(150L); } catch(Exception e) { }
+        try { Thread.sleep(150L); } catch(InterruptedException e) { }
         assertTrue("Should be 10 idle, found " + pool.getNumIdle(),pool.getNumIdle() == 10);
 
         for(int i=0 ; i<10 ; i++) {
             active[i] = pool.borrowObject();
         }
 
-        try { Thread.sleep(150L); } catch(Exception e) { }
+        try { Thread.sleep(150L); } catch(InterruptedException e) { }
         assertTrue("Should be 0 idle, found " + pool.getNumIdle(),pool.getNumIdle() == 0);
 
         for(int i=0 ; i<10 ; i++) {
             pool.returnObject(active[i]);
         }
 
-        try { Thread.sleep(150L); } catch(Exception e) { }
+        try { Thread.sleep(150L); } catch(InterruptedException e) { }
         assertTrue("Should be 10 idle, found " + pool.getNumIdle(),pool.getNumIdle() == 10);
     }
 
@@ -941,12 +1093,12 @@ public class TestGenericObjectPool extends TestBaseObjectPool {
             while(!(threads[i]).complete()) {
                 try {
                     Thread.sleep(500L);
-                } catch(Exception e) {
+                } catch(InterruptedException e) {
                     // ignored
                 }
             }
             if(threads[i].failed()) {
-                fail();
+                fail("Thread "+i+" failed: "+threads[i]._error.toString());
             }
         }
     }
@@ -978,36 +1130,44 @@ public class TestGenericObjectPool extends TestBaseObjectPool {
         runTestThreads(5, 10, 50);
     }
 
-    class TestThread implements Runnable {
-        java.util.Random _random = new java.util.Random();
-        ObjectPool _pool = null;
-        boolean _complete = false;
-        boolean _failed = false;
-        int _iter = 100;
-        int _delay = 50;
-        boolean _randomDelay = true;
+    static class TestThread implements Runnable {
+        private final java.util.Random _random = new java.util.Random();
+        
+        // Thread config items
+        private final ObjectPool _pool;
+        private final int _iter;
+        private final int _delay;
+        private final boolean _randomDelay;
+        private final Object _expectedObject;
+        
+        private volatile boolean _complete = false;
+        private volatile boolean _failed = false;
+        private volatile Throwable _error;
 
         public TestThread(ObjectPool pool) {
-            _pool = pool;
+            this(pool, 100, 50, true, null);
         }
 
         public TestThread(ObjectPool pool, int iter) {
-            _pool = pool;
-            _iter = iter;
+            this(pool, iter, 50, true, null);
         }
 
         public TestThread(ObjectPool pool, int iter, int delay) {
-            _pool = pool;
-            _iter = iter;
-            _delay = delay;
+            this(pool, iter, delay, true, null);
         }
         
         public TestThread(ObjectPool pool, int iter, int delay,
                 boolean randomDelay) {
+            this(pool, iter, delay, randomDelay, null);
+        }
+
+        public TestThread(ObjectPool pool, int iter, int delay,
+                boolean randomDelay, Object obj) {
             _pool = pool;
             _iter = iter;
             _delay = delay;
             _randomDelay = randomDelay;
+            _expectedObject = obj;
         }
 
         public boolean complete() {
@@ -1024,27 +1184,35 @@ public class TestGenericObjectPool extends TestBaseObjectPool {
                     _randomDelay ? (long)_random.nextInt(_delay) : _delay;
                 try {
                     Thread.sleep(delay);
-                } catch(Exception e) {
+                } catch(InterruptedException e) {
                     // ignored
                 }
                 Object obj = null;
                 try {
                     obj = _pool.borrowObject();
                 } catch(Exception e) {
-                    e.printStackTrace();
+                    _error = e;
                     _failed = true;
                     _complete = true;
                     break;
                 }
 
+                if (_expectedObject != null && !_expectedObject.equals(obj)) {
+                    _error = new Throwable("Expected: "+_expectedObject+ " found: "+obj);
+                    _failed = true;
+                    _complete = true;
+                    break;
+                }
+                
                 try {
                     Thread.sleep(delay);
-                } catch(Exception e) {
+                } catch(InterruptedException e) {
                     // ignored
                 }
                 try {
                     _pool.returnObject(obj);
                 } catch(Exception e) {
+                    _error = e;
                     _failed = true;
                     _complete = true;
                     break;
@@ -1084,15 +1252,15 @@ public class TestGenericObjectPool extends TestBaseObjectPool {
 
     public void testAddObject() throws Exception {
         assertEquals("should be zero idle", 0, pool.getNumIdle());
-    	pool.addObject();
-		assertEquals("should be one idle", 1, pool.getNumIdle());
-		assertEquals("should be zero active", 0, pool.getNumActive());
-		Object obj = pool.borrowObject();
-		assertEquals("should be zero idle", 0, pool.getNumIdle());
-		assertEquals("should be one active", 1, pool.getNumActive());
-		pool.returnObject(obj);
-		assertEquals("should be one idle", 1, pool.getNumIdle());
-		assertEquals("should be zero active", 0, pool.getNumActive());
+        pool.addObject();
+        assertEquals("should be one idle", 1, pool.getNumIdle());
+        assertEquals("should be zero active", 0, pool.getNumActive());
+        Object obj = pool.borrowObject();
+        assertEquals("should be zero idle", 0, pool.getNumIdle());
+        assertEquals("should be one active", 1, pool.getNumActive());
+        pool.returnObject(obj);
+        assertEquals("should be one idle", 1, pool.getNumIdle());
+        assertEquals("should be zero active", 0, pool.getNumActive());
 
         ObjectPool op = new GenericObjectPool();
         try {
@@ -1165,13 +1333,16 @@ public class TestGenericObjectPool extends TestBaseObjectPool {
             }
             return String.valueOf(makeCounter++);
         }
-        public void destroyObject(Object obj) {
+        public void destroyObject(Object obj) throws Exception {
             if (destroyLatency > 0) {
                 doWait(destroyLatency);
             }
             synchronized(this) {
                 activeCount--;
             }
+            if (exceptionOnDestroy) {
+                throw new Exception();
+            }
         }
         public boolean validateObject(Object obj) {
             if (enableValidation) { 
@@ -1200,6 +1371,7 @@ public class TestGenericObjectPool extends TestBaseObjectPool {
         boolean oddValid = true;
         boolean exceptionOnPassivate = false;
         boolean exceptionOnActivate = false;
+        boolean exceptionOnDestroy = false;
         boolean enableValidation = true;
         long destroyLatency = 0;
         long makeLatency = 0;
@@ -1212,6 +1384,10 @@ public class TestGenericObjectPool extends TestBaseObjectPool {
         public void setThrowExceptionOnActivate(boolean b) {
             exceptionOnActivate = b;
         }
+        
+        public void setThrowExceptionOnDestroy(boolean b) {
+            exceptionOnDestroy = b;
+        }
 
         public boolean isValidationEnabled() {
             return enableValidation;
@@ -1236,6 +1412,50 @@ public class TestGenericObjectPool extends TestBaseObjectPool {
     protected boolean isFifo() {
         return false;
     }
+
+    /*
+     * Note: This test relies on timing for correct execution. There *should* be
+     * enough margin for this to work correctly on most (all?) systems but be
+     * aware of this if you see a failure of this test.
+     */
+    public void testBorrowObjectFairness() {
+        // Config
+        int numThreads = 30;
+        int maxActive = 10;
+
+        SimpleFactory factory = new SimpleFactory();
+        factory.setMaxActive(maxActive);
+        pool.setFactory(factory);
+        pool.setMaxActive(maxActive);
+        pool.setWhenExhaustedAction(GenericObjectPool.WHEN_EXHAUSTED_BLOCK);
+        pool.setTimeBetweenEvictionRunsMillis(-1);
+
+        // Start threads to borrow objects
+        TestThread[] threads = new TestThread[numThreads];
+        for(int i=0;i<numThreads;i++) {
+            threads[i] = new TestThread(pool, 1, 2000, false, String.valueOf(i % maxActive));
+            Thread t = new Thread(threads[i]);
+            t.start();
+            // Short delay to ensure threads start in correct order
+            try {
+                Thread.sleep(50);
+            } catch (InterruptedException e) {
+                fail(e.toString());
+            }
+        }
+
+        // Wait for threads to finish
+        for(int i=0;i<numThreads;i++) {
+            while(!(threads[i]).complete()) {
+                try {
+                    Thread.sleep(500L);
+                } catch(InterruptedException e) {
+                    // ignored
+                }
+            }
+            if(threads[i].failed()) {
+                fail("Thread "+i+" failed: "+threads[i]._error.toString());
+            }
+        }
+    }
 }
-
-
